@@ -6,14 +6,22 @@ from pathlib import Path
 from functools import partial
 from typing import Optional, Dict, List, Tuple, Union, Any
 import time
+from collections import Counter
 
 import numpy as np
 import torch
+import hashlib
+import portalocker
+import tqdm
 
-from rtg import log, yaml
+import rtg
+from rtg import log, yaml, device
 from rtg.data.dataset import (TSVData, BatchIterable, LoopingIterable, SqliteFile, GenerativeBatchIterable)
 from rtg.data.codec import Field, SPField, NLField, PretrainMatchField
 from rtg.utils import IO, line_count
+from rtg.registry import CRITERION, OPTIMIZER, SCHEDULE, MODEL
+from rtg.schema import config_checks
+
 
 seeded = False
 
@@ -41,6 +49,7 @@ class BaseExperiment:
         if isinstance(config, str) or isinstance(config, Path):
             config = load_conf(config)
         self.config = config if config else load_conf(self._config_file)
+        config_checks(self.config)
         self.codec_name = self.config.get('prep', {}).get('codec_lib', 'sentpiece')  # with default
         codec_libs = {'sentpiece': SPField,
                       'nlcodec': NLField,
@@ -74,6 +83,10 @@ class BaseExperiment:
         self.shared_field = self.Field(str(self._shared_field_file)) \
             if self._shared_field_file.exists() else None
 
+    @property
+    def problem_type(self):
+        raise NotImplementedError
+
     def maybe_seed(self):
         global seeded
         if not seeded and 'seed' in self.config:
@@ -106,11 +119,11 @@ class BaseExperiment:
     def has_trained(self):
         return self._trained_flag.exists()
 
-    def store_model(self, epoch: int, model, train_score: float, val_score: float, keep: int,
+    def store_model(self, optimizer_step: int, model, train_score: float, val_score: float, keep: int,
                     prefix='model', keeper_sort='step'):
         """
         saves model to a given path
-        :param epoch: epoch number of model
+        :param optimizer_step: optimizer step of the model
         :param model: model object itself
         :param train_score: score of model on training split
         :param val_score: score of model on validation split
@@ -124,9 +137,9 @@ class BaseExperiment:
         if self.read_only:
             log.warning("Ignoring the store request; experiment is readonly")
             return
-        name = f'{prefix}_{epoch:03d}_{train_score:.6f}_{val_score:.6f}.pkl'
+        name = f'{prefix}_{optimizer_step:03d}_{train_score:.6f}_{val_score:.6f}.pkl'
         path = self.model_dir / name
-        log.info(f"Saving epoch {epoch} to {path}")
+        log.info(f"Saving optimizer step {optimizer_step} to {path}")
         torch.save(model, str(path))
 
         del_models = []
@@ -141,7 +154,7 @@ class BaseExperiment:
             os.remove(str(d_model))
 
         with IO.writer(os.path.join(self.model_dir, 'scores.tsv'), append=True) as f:
-            cols = [str(epoch), datetime.now().isoformat(), name, f'{train_score:g}',
+            cols = [str(optimizer_step), datetime.now().isoformat(), name, f'{train_score:g}',
                     f'{val_score:g}']
             f.write('\t'.join(cols) + '\n')
 
@@ -195,7 +208,6 @@ class BaseExperiment:
         :return: Tuple[Optional[Path], step_num:int]
         """
         models = self.list_models(sort=sort, desc=desc)
-        print(models)
         if models:
             name = models[0].name.replace('.pkl', '').replace('.txt.gz', '')
             step, train_score, valid_score = name.split('_')[-3:]
@@ -226,26 +238,6 @@ class BaseExperiment:
         """
         self.config['model_args'] = model_args
 
-    @property
-    def optim_args(self) -> Tuple[Optional[str], Dict]:
-        """
-        Gets optimizer args from file
-        :return: optimizer args if exists or None otherwise
-        """
-        opt_conf = self.config.get('optim')
-        if opt_conf:
-            return opt_conf.get('name'), opt_conf.get('args')
-        else:
-            return None, {}
-
-    @optim_args.setter
-    def optim_args(self, optim_args: Tuple[str, Dict]):
-        """
-        set optimizer args
-        """
-        name, args = optim_args
-        self.config['optim'].update({'name': name, 'args': args})
-
     @property
     def shared_vocab(self) -> Field:
         return self.shared_field
@@ -259,6 +251,131 @@ class BaseExperiment:
                 return p
         return None
 
+    def pre_process(self, args=None, force=False):
+        if self.has_prepared() and not force:
+            log.warning("Already prepared")
+            return
+        args = args if args else self.config['prep']
+        if 'parent' in self.config:
+            self.inherit_parent()
+
+        if 'same_data' in args:
+            data = Path(args['same_data']) / 'data'
+            assert data.exists()
+            log.info(f"Reusing prepared data dir from {data}")
+            if self.data_dir.exists():
+                if self.data_dir.is_symlink():
+                    self.data_dir.unlink()
+                else:
+                    self.data_dir.rename('data.bak')
+            self.data_dir.symlink_to(data.resolve(), target_is_directory=True)
+            self.reload()
+            self._prepared_flag.touch()
+
+    def inherit_parent(self):
+        raise NotImplemented()
+
+    def train(self, args=None):
+        raise NotImplementedError()
+
+    def reload(self):
+        exp = type(self)(self.work_dir, read_only=self.read_only)
+        self.__dict__ = exp.__dict__
+
+    @classmethod
+    def _checkpt_to_model_state(cls, checkpt_path: Union[str, Path]):
+        state = torch.load(checkpt_path, map_location=device)
+        if 'model_state' in state:
+            state = state['model_state']
+        return state
+
+    @classmethod
+    def average_states(cls, model_paths: List[Path]):
+        assert model_paths, 'at least one model checkpoint should be given. Check your directory'
+        for i, mp in enumerate(model_paths):
+            next_state = cls._checkpt_to_model_state(mp)
+            if i < 1:
+                state_dict = next_state
+                key_set = set(state_dict.keys())
+            else:
+                # noinspection PyUnboundLocalVariable
+                assert key_set == set(next_state.keys())
+                for key in key_set:     # Running average
+                    state_dict[key] = (i*state_dict[key] + next_state[key]) / (i + 1)
+        return state_dict
+
+    def maybe_ensemble_state(self, model_paths: Optional[List[str]], ensemble: int = 1):
+        if model_paths and len(model_paths) == 1:
+            log.info(f" Restoring state from requested model {model_paths[0]}")
+            return self._checkpt_to_model_state(model_paths[0])
+        elif not model_paths and ensemble <= 1:
+            model_path, _ = self.get_best_known_model()
+            log.info(f" Restoring state from best known model: {model_path}")
+            return self._checkpt_to_model_state(model_path)
+        else:
+            if not model_paths:
+                # Average last n models
+                model_paths = self.list_models(sort='step', desc=True)[:ensemble]
+            digest = hashlib.md5(";".join(str(p) for p in model_paths).encode('utf-8')).hexdigest()
+            cache_file = self.model_dir / f'avg_state{len(model_paths)}_{digest}.pkl'
+            lock_file = cache_file.with_suffix('.lock')
+            MAX_TIMEOUT = 1 * 60 * 60  # 1 hour
+            with portalocker.Lock(lock_file, 'w', timeout=MAX_TIMEOUT) as fh:
+                # check if downloaded by  other parallel process
+                if lock_file.exists() and cache_file.exists():
+                    log.info(f"Cache exists: reading from {cache_file}")
+                    state = self._checkpt_to_model_state(cache_file)
+                else:
+                    log.info(f"Averaging {len(model_paths)} model states :: {model_paths}")
+                    state = self.average_states(model_paths)
+                    if len(model_paths) > 1:
+                        log.info(f"Caching the averaged state at {cache_file}")
+                        torch.save(state, str(cache_file))
+            return state
+        
+    def load_model(self, model_paths=None, ensemble=1):
+        from .registry import MODELS
+        model = MODELS[self.model_type].Model(exp=self, **self.model_args)[0]
+        state = self.maybe_ensemble_state(model_paths=model_paths, ensemble=ensemble)
+        errors = model.load_state_dict(state)
+        log.info(f"{errors}")
+        return model
+
+    def load_model_with_state(self, checkpt_state):
+        from .registry import MODELS
+        chkpt = checkpt_state
+        state = chkpt['model_state']
+        model_type = chkpt['model_type']
+        model_args = chkpt['model_args']
+        model = MODELS[model_type].Model(exp=self, **model_args)[0]
+        errors = model.load_state_dict(state)
+        log.info(f"{errors}")
+        log.info(f"Successfully restored the model state of : {model_type}")
+        return model
+
+    def get_conf_component(self, kind, extra_args=None):
+        """ Creates a component such as schedule, criterion, optimizer based on config"""
+        from rtg.registry import registry
+        assert kind in registry, f'component {kind} is unknown; valid: {registry.keys()}'
+        if not kind in self.config:
+            log.warning(f"{kind} not found in config; skipping")
+            return None
+        name, args = self.config[kind]['name'], self.config[kind].get('args') or {}
+        assert name in registry[kind], f'{kind}={name} is invalid; valid: {registry[kind].keys()}'
+        factory = registry[kind][name]
+        extra_args = extra_args or {}
+        log.info(f"creating {kind} {name} with args {args}")
+        return factory(**extra_args, **args)
+
+    def get_criterion(self, extra_args=None):
+        return self.get_conf_component(CRITERION, extra_args=extra_args)
+
+    def get_schedule(self):
+        return self.get_conf_component(SCHEDULE)
+
+    def get_optimizer(self, params):
+        return self.get_conf_component(OPTIMIZER, extra_args=dict(params=params))
+
 
 class TranslationExperiment(BaseExperiment):
 
@@ -272,7 +389,8 @@ class TranslationExperiment(BaseExperiment):
         self.emb_tgt_file = self.data_dir / 'emb_tgt.pt'
         self.ext_emb_src_file = self.data_dir / 'ext_emb_src.pt'  # external Embeddings
         self.ext_emb_tgt_file = self.data_dir / 'ext_emb_tgt.pt'  # external Embeddings
-
+        self.src_field = None
+        self.tgt_field = None
         self.reload_vocabs()
 
         # Either shared field  OR  individual  src and tgt fields
@@ -290,6 +408,11 @@ class TranslationExperiment(BaseExperiment):
 
         self.parent_model_state = self.data_dir / 'parent_model_state.pt'
 
+    @property
+    def problem_type(self):
+        from rtg.registry import ProblemType
+        return ProblemType.TRANSLATION
+
     def reload_vocabs(self):
         self.src_field, self.tgt_field, self.shared_field = [
             self.Field(str(f)) if f.exists() else None for f in (
@@ -315,23 +438,38 @@ class TranslationExperiment(BaseExperiment):
 
         xt_args = dict(no_split_toks=args.get('no_split_toks'),
                        char_coverage=args.get('char_coverage', 0))
+        min_co_ev = args.get('min_co_ev', None)
+        pieces = args['pieces']
         if args.get('shared_vocab'):  # shared vocab
+            assert 'max_types' in args, f'prep.max_types is required when prep.shared_vocab=true'
+            max_types = args['max_types']
             corpus = [args[key] for key in ['train_src', 'train_tgt', 'mono_src', 'mono_tgt']
                       if args.get(key)]
-            self.shared_field = self._make_vocab("shared", self._shared_field_file, args['pieces'],
-                                                 args['max_types'], corpus=corpus, **xt_args)
+            assert isinstance(pieces, str), f'shared vocab cant support different pieces for src, tgt;' \
+                                            f' given pieces={pieces}. Either set shared=false or pieces=<a string>'
+            self.shared_field = self._make_vocab("shared", self._shared_field_file, pieces, max_types,
+                                                 corpus=corpus, min_co_ev=min_co_ev, **xt_args)
         else:  # separate vocabularies
             src_corpus = [args[key] for key in ['train_src', 'mono_src'] if args.get(key)]
-            self.src_field = self._make_vocab("src", self._src_field_file, args['pieces'],
-                                              args['max_src_types'], corpus=src_corpus, **xt_args)
+            src_min_co_ev = args.get('src_min_co_ev', min_co_ev)
+            tgt_min_co_ev = args.get('tgt_min_co_ev', min_co_ev)
+            src_pieces = tgt_pieces = pieces
+            if not isinstance(pieces, str):
+                assert len(pieces) == 2
+                src_pieces, tgt_pieces = pieces
+                log.info(f"Vocab types: src: {src_pieces} and tgt: {tgt_pieces}")
 
+            max_src_types = args.get('max_src_types', args.get('max_types'))
+            max_tgt_types = args.get('max_tgt_types', args.get('max_types'))
+            assert max_src_types and max_tgt_types, 'prep.{max_src_types,max_tgt_types} are required' \
+                                                    ' when prep.shared_vocab=false'
+            self.src_field = self._make_vocab("src", self._src_field_file, src_pieces, max_src_types, corpus=src_corpus,
+                                              min_co_ev=src_min_co_ev, **xt_args)
             # target vocabulary
             tgt_corpus = [args[key] for key in ['train_tgt', 'mono_tgt'] if args.get(key)]
-            self.tgt_field = self._make_vocab("src", self._tgt_field_file, args['pieces'],
-                                              args['max_tgt_types'], corpus=tgt_corpus, **xt_args)
-
+            self.tgt_field = self._make_vocab("src", self._tgt_field_file, tgt_pieces, max_tgt_types, corpus=tgt_corpus,
+                                              min_co_ev=tgt_min_co_ev, **xt_args)
         train_file = self.train_db
-
         self._pre_process_parallel('train_src', 'train_tgt', out_file=train_file, args=args,
                                    line_check=False)
         self._pre_process_parallel('valid_src', 'valid_tgt', out_file=self.valid_file, args=args,
@@ -352,7 +490,8 @@ class TranslationExperiment(BaseExperiment):
         TSVData.write_parallel_recs(samples, self.samples_file)
 
     def _make_vocab(self, name: str, vocab_file: Path, model_type: str, vocab_size: int,
-                    corpus: List, no_split_toks: List[str] = None, char_coverage=0) -> Field:
+                    corpus: List, no_split_toks: List[str] = None, char_coverage=0,
+                    min_co_ev=None) -> Field:
         """
         Construct vocabulary file
         :param name: name : src, tgt or shared -- for the sake of logging
@@ -375,8 +514,11 @@ class TranslationExperiment(BaseExperiment):
 
         flat_uniq_corpus = list(flat_uniq_corpus)
         log.info(f"Going to build {name} vocab from files")
+        xt_args = {}
+        if min_co_ev:
+            xt_args["min_co_ev"] = min_co_ev
         return self.Field.train(model_type, vocab_size, str(vocab_file), flat_uniq_corpus,
-                                no_split_toks=no_split_toks, char_coverage=char_coverage)
+                                no_split_toks=no_split_toks, char_coverage=char_coverage, **xt_args)
 
     def pre_process_mono(self, args):
         xt_args = dict(no_split_toks=args.get('no_split_toks'),
@@ -453,7 +595,10 @@ class TranslationExperiment(BaseExperiment):
             args[src_key], args[tgt_key], args['truncate'], args['src_len'], args['tgt_len'],
             src_tokenizer=partial(self.src_vocab.encode_as_ids, split_ratio=split_ratio),
             tgt_tokenizer=partial(self.tgt_vocab.encode_as_ids, split_ratio=split_ratio))
-        if any([out_file.name.endswith(suf) for suf in ('.db', '.db.tmp')]):
+        if any([out_file.name.endswith(suf) for suf in ('.nldb', '.nldb.tmp')]):
+            from nlcodec.db import MultipartDb
+            MultipartDb.create(path=out_file, recs=parallel_recs, field_names=('x', 'y'))
+        elif any([out_file.name.endswith(suf) for suf in ('.db', '.db.tmp')]):
             SqliteFile.write(out_file, records=parallel_recs)
         else:
             TSVData.write_parallel_recs(parallel_recs, out_file)
@@ -591,7 +736,7 @@ class TranslationExperiment(BaseExperiment):
 
     def inherit_parent(self):
         parent = self.config['parent']
-        parent_exp = TranslationExperiment(parent['experiment'], read_only=True)
+        parent_exp = type(self)(parent['experiment'], read_only=True)
         log.info(f"Parent experiment: {parent_exp.work_dir}")
         parent_exp.has_prepared()
         vocab_sepc = parent.get('vocab')
@@ -626,8 +771,7 @@ class TranslationExperiment(BaseExperiment):
             ensemble = model_sepc.get('ensemble', 1)
             model_paths = parent_exp.list_models(sort='step', desc=True)[:ensemble]
             log.info(f"Averaging {len(model_paths)} checkpoints of parent model: \n{model_paths}")
-            from rtg.module.decoder import Decoder
-            avg_state = Decoder.average_states(model_paths=model_paths)
+            avg_state = self.average_states(model_paths=model_paths)
             log.info(f"Saving parent model's state to {self.parent_model_state}")
             torch.save(avg_state, self.parent_model_state)
 
@@ -662,38 +806,17 @@ class TranslationExperiment(BaseExperiment):
             torch.save(avg_state, self.parent_model_state)
             self.persist_state()  # this will fix src_vocab and tgt_vocab of model_args conf
 
-
     def pre_process(self, args=None, force=False):
+        args = args or self.config['prep']
+        super(TranslationExperiment, self).pre_process(args, )
         if self.has_prepared() and not force:
             log.warning("Already prepared")
             return
-        args = args if args else self.config['prep']
-        if 'parent' in self.config:
-            self.inherit_parent()
 
-        if 'same_data' in args:
-            data = Path(args['same_data']) / 'data'
-            assert data.exists()
-            log.info(f"Reusing prepared data dir from {data}")
-            if self.data_dir.exists():
-                if self.data_dir.is_symlink():
-                    self.data_dir.unlink()
-                else:
-                    self.data_dir.rename('data.bak')
-            self.data_dir.symlink_to(data.resolve(), target_is_directory=True)
-            self.reload_vocabs()
+        if self._unsupervised:
+            self.pre_process_mono(args)
         else:
-            vocabs = args.get('vocabs')
-            if vocabs:
-                parent = TranslationExperiment(vocabs, read_only=True)
-                parent.copy_vocabs(self)
-                self.shared_field, self.src_field, self.tgt_field = [
-                    self.Field(str(f)) if f.exists() else None
-                    for f in (self._shared_field_file, self._src_field_file, self._tgt_field_file)]
-            if self._unsupervised:
-                self.pre_process_mono(args)
-            else:
-                self.pre_process_parallel(args)
+            self.pre_process_parallel(args)
 
         self.maybe_pre_process_embeds()
         # update state on disk
@@ -717,6 +840,12 @@ class TranslationExperiment(BaseExperiment):
             args['tgt_vocab'] = len(self.tgt_vocab) if self.tgt_vocab else 0
 
         self.config['updated_at'] = datetime.now().isoformat()
+        if 'rtg_version' not in self.config:
+            self.config['rtg_version'] = {}
+        version = self.config['rtg_version']
+        if version.get('last_worked', None) != rtg.__version__:
+            version['previous'] =  version.get('last_worked')
+            version['last_worked'] = rtg.__version__
         self.store_config()
 
     def train(self, args=None):
@@ -745,11 +874,8 @@ class TranslationExperiment(BaseExperiment):
             log.warning(
                 f"Already trained upto {last_step}; Requested: train={train_steps}, finetune={finetune_steps} Skipped")
             return
-
-        from rtg.registry import trainers, factories
-        name, optim_args = self.optim_args
-        trainer = trainers[self.model_type](self, optim=name,
-                                            model_factory=factories[self.model_type], **optim_args)
+        from .registry import MODELS
+        trainer = MODELS[self.model_type].Trainer(self)
         if last_step < train_steps:  # regular training
             stopped = trainer.train(fine_tune=False, **run_args)
             if not self.read_only:
@@ -787,9 +913,9 @@ class TranslationExperiment(BaseExperiment):
                 [('src_len', 'max_src_len'), ('tgt_len', 'max_tgt_len'), ('truncate', 'truncate')]
                 if ik in prep_args}
 
-    def get_train_data(self, batch_size:  Union[int, Tuple[int,int]], steps: int = 0, sort_by='eq_len_rand_batch',
+    def get_train_data(self, batch_size:  Union[int, Tuple[int, int]], steps: int = 0, sort_by='eq_len_rand_batch',
                        batch_first=True, shuffle=False, fine_tune=False, keep_in_mem=False,
-                       split_ratio: float = 0., dynamic_epoch=False):
+                       split_ratio: float = 0., dynamic_epoch=False, y_is_cls=False):
 
         data_path = self.train_db if self.train_db.exists() else self.train_file
         if fine_tune:
@@ -802,35 +928,41 @@ class TranslationExperiment(BaseExperiment):
         if split_ratio > 0:
             data_path = IO.maybe_tmpfs(data_path)
             train_file = data_path.with_suffix('.db.tmp')
+            assert not y_is_cls, 'Not supported feature'
             file_creator = partial(self.file_creator, train_file=train_file, split_ratio=split_ratio)
             train_data = GenerativeBatchIterable(
                 file_creator=file_creator, batches=steps, batch_size=batch_size, field=self.tgt_vocab,
                 dynamic_epoch=dynamic_epoch, batch_first=batch_first, shuffle=shuffle, sort_by=sort_by,
                 **self._get_batch_args())
         else:
-            data = BatchIterable(
+            train_data = BatchIterable(
                 data_path=data_path, batch_size=batch_size, field=self.tgt_vocab, sort_by=sort_by,
-                batch_first=batch_first, shuffle=shuffle, **self._get_batch_args())
-            train_data = LoopingIterable(data, steps)
-
+                batch_first=batch_first, shuffle=shuffle, y_is_cls=y_is_cls, **self._get_batch_args())
+            # default, read data once completely, if steps > 0, truncate or loop depending on steps and data size
+            if steps > 0:
+                train_data = LoopingIterable(train_data, steps)
         return train_data
 
-
     def file_creator(self, train_file, split_ratio, *args, **kwargs):
         self._pre_process_parallel(*args, src_key='train_src', tgt_key='train_tgt',
                                    out_file=train_file, split_ratio=split_ratio, **kwargs)
         return train_file
 
-    def get_val_data(self, batch_size: Union[int, Tuple[int,int]], sort_desc=False, batch_first=True,
-                     shuffle=False):
-        raw_path = None
+    def get_val_data(self, batch_size: Union[int, Tuple[int, int]], sort_desc=False, batch_first=True,
+                     shuffle=False, y_is_cls=False):
         prep = self.config.get('prep', {})
-        if 'valid_src' in prep and 'valid_tgt' in prep:
-            raw_path = prep['valid_src'], prep['valid_tgt']
-
+        raw_tgt = prep.get('valid_tgt_raw', None)
+        if not raw_tgt:
+            raise Exception('Config value prep.valid_tgt_raw is required. It should have path to a file'
+                            ' having raw (unmodified) target file, to be used for computing BLEU.')
+        raw_src = prep.get('valid_src_raw', prep.get('valid_src'))
+        for path in (raw_src, raw_tgt):
+            assert Path(path).exists(), f'File at {path} does not exist; it is required'
+        raw_path = Path(raw_src), Path(raw_tgt)
         return BatchIterable(self.valid_file, batch_size=batch_size, sort_desc=sort_desc,
                              batch_first=batch_first, shuffle=shuffle, field=self.tgt_vocab,
-                             keep_in_mem=True, raw_path=raw_path, **self._get_batch_args())
+                             keep_in_mem=True, raw_path=raw_path, y_is_cls=y_is_cls,
+                             **self._get_batch_args())
 
     def get_combo_data(self, batch_size: int, steps: int = 0, sort_desc=False, batch_first=True,
                        shuffle=False):
@@ -840,16 +972,11 @@ class TranslationExperiment(BaseExperiment):
         combo_file = IO.maybe_tmpfs(self.combo_file)
         data = BatchIterable(
             combo_file, batch_size=batch_size, sort_desc=sort_desc, field=self.tgt_vocab,
-            batch_first=batch_first, shuffle=shuffle, **self._get_batch_args()
-        )
+            batch_first=batch_first, shuffle=shuffle, **self._get_batch_args())
         if steps > 0:
             data = LoopingIterable(data, steps)
         return data
 
-    def reload(self):
-        exp = type(self)(self.work_dir, read_only=self.read_only)
-        self.__dict__ = exp.__dict__
-
     def copy_vocabs(self, other):
         """
         Copies vocabulary files from self to other
@@ -902,3 +1029,48 @@ class TranslationExperiment(BaseExperiment):
         if num_batches > 0:
             data = LoopingIterable(data, num_batches)
         return data
+
+    def get_class_freqs(self):
+        batch_size = self.config.get('trainer', {}).get('batch_size', 2000)
+        train_data = self.get_train_data(batch_size=batch_size, steps=-1, shuffle=False)
+        freq_file = self.data_dir / 'class.freqs.tsv'
+        vocab = self.tgt_vocab
+
+        if not freq_file.exists():
+            stats = Counter()
+            for batch in tqdm.tqdm(train_data):
+                for seq in batch.y_seqs.tolist():
+                    stats.update(seq)
+            if vocab.pad_idx >= 0:
+                stats[vocab.pad_idx] = 0
+
+            with freq_file.open('w', encoding='utf-8') as out:
+                for i in range(len(vocab)):
+                    name = vocab.class_names[i]
+                    freq = stats.get(i, 0)
+                    out.write(f'{i}\t{name}\t{freq}\n')
+
+        with freq_file.open('r') as lines:
+            recs = (line.split('\t') for line in lines)
+            recs = [(int(r[0]), r[1], int(r[2])) for r in recs]
+            return recs
+
+    def get_pre_transform(self, side: str):
+        assert side in ('src', 'tgt')
+        from rtg.transform import TextTransform
+        conf_chain = self.config.get('prep', {}).get(f'{side}_pre_proc', None)
+        if conf_chain:
+            transform = TextTransform.make(names=conf_chain)
+        else:
+            transform = TextTransform.recommended_pre()
+        return transform
+
+    def get_post_transform(self, side: str):
+        assert side in ('src', 'tgt')
+        from rtg.transform import TextTransform
+        conf_chain = self.config.get('prep', {}).get(f'{side}_post_proc', None)
+        if conf_chain:
+            transform = TextTransform.make(names=conf_chain)
+        else:
+            transform = TextTransform.recommended_post()
+        return transform
