@@ -6,8 +6,4 @@ from rtg.module import Model
 from abc import ABCMeta, ABC
 """
 Stuff related to Language Model
-"""
-
-
-class LanguageModel(Model, metaclass=ABCMeta):
-    pass
+"""
\ No newline at end of file
